require_relative "./promise.rb"

def zipn(n, a, f)
  return f[*a] if n <= 0 || a==[]
  return [] if a.any?{|i|i.empty}
  [promise{zipn(n-1, a.map{|i|i.value[0]}, f) },
   promise{zipn(n, a.map{|i|i.value[1]}, f) }]
end

class Promise
  def initialize(&block)
    @impl=block
  end
  def value
    @impl=@impl[] if Proc===@impl
    @impl
  end
  def empty
    value.empty?
  end
end

def promise(&b)
  Promise.new(&b)
end

def concat_map(v,rhs,first=true,&b)
  if v.empty
    rhs.value
  else
    b[v.value[0],Promise.new{concat_map(v.value[1],rhs,false,&b)},first]
  end
end

def append(v,r)
  v.empty ? r.value : [v.value[0],promise{append(v.value[1], r)}]
end

def join(a,b)
  concat_map(a,Null){|i,r,first|
    first ? append(i,r) : append(b,promise{append(i,r)})
  }
end

def starts_with(a,b) # assumed to be strs, returns remainder if match otherwise nil
  return a if b.empty
  return nil if a.empty || a.value[0].value != b.value[0].value
  starts_with(a.value[1],b.value[1])
end

def split(a,b) # fyi could be lazier since we know we return a list always
  return [Null,Null] if a.empty
  if (remainder=starts_with(a,b))
    [Null,Promise.new{split(remainder,b)}]
  else
    rhs=Promise.new{split(a.value[1],b)}
    [promise{[a.value[0],Promise.new{rhs.value[0].value}]},promise{rhs.value[1].value}]
  end
end

class Object
  def const
    Promise.new{self}
  end
end

def repeat(a)
  (ret = [a]) << ret.const
end

def repn(n, a)
  n<=0 ? a : promise{ repeat(repn(n-1, a)) }
end

def promoten(n,a)
  if n<=0
    a
  else
    [promoten(n-1,a),Null].const
  end
end

def last(a, prev)
  until a.empty
    prev = a.value[0]
    a = a.value[1]
  end
  prev.value
end

def init(a)
  return [] if a.empty || a.value[1].empty
  [a.value[0], Promise.new{ init(a.value[1]) }]
end

FalseChars = {0=>1,9=>1,10=>1,11=>1,12=>1,13=>1,32=>1}
def truthy(type, rank, value)
  return !value.empty? if rank > 0
  case type
    when IntType
      return value != 0
    when CharType
      !FalseChars.include?(value)
    else raise
  end
end

def reverse(a,sofar=[])
  return sofar if a.empty
  reverse(a.value[1],[a.value[0],sofar.const])
end

def chunk_while(a,b,rank,btype)
  raise IogiiError.new "cannot chunk_while negative rank" if rank < 0 # todo ideally this done statically
  return [Null,Null] if a.empty || b.empty
  b0true = Promise.new{ truthy(btype,rank,b.value[0].value)}
  rhs = Promise.new{ chunk_while(a.value[1],b.value[1],rank,btype) }
  [
    Promise.new{
      if b0true.value
        [a.value[0], rhs.value[0]]
      else
        []
      end
    },
    Promise.new{
      if b0true.value
        rhs.value[1].value
      else
        rhs.value
      end
    }
  ]
end

def superpad(n, a)
  if n > 0
    pad(promise { map(a){|e| superpad(n-1, e) } } )
  else
    a.value
  end
end

def pad(a)
  [
    promise{ a.empty ? raise(IogiiError.new("pad value used (see #{NittyGritty}#superpad )")) : a.value[0].value },
    promise{ pad( promise{ a.empty ? [] : a.value[1].value }) }
  ]
end

def padv(a,b)
  [
    promise{ a.empty ? b.value : a.value[0].value },
    promise{ padv( promise{ a.empty ? [] : a.value[1].value }, b) }
  ]
end

def printv(a, type, rank)
  srank = rank + (type == IntType ? 1 : 0)
  n = [srank == 2 ? 1 : srank - 2, 0].max
  printn(a, n, type, srank - n)
end

def printn(a, n, type, srank)
  if n == 0
    if srank == 0
      prints([a.const, Null])
    elsif srank == 1
      case type
      when EmptyType

      when IntType
        print a
      when CharType
        prints a
      else raise
      end
    elsif srank == 2
      until a.empty?
        printn(a[0].value, 0, type, 1)
        print " "
        a = a[1].value
      end
    else
      raise
    end
  else
    until a.empty?
      printn(a[0].value, n-1, type, srank)
      puts
      a = a[1].value
    end
  end
end

def prints(value)
  while !value.empty?
    v = value[0].value
    # raw print, multiple bytes will show up as one if utf is used. output will match source code so will look correct if editor encoding=terminal encoding
    raise IogiiError.new("char value, %s, outside of byte range" % v, nil) if v<0 || v>255
    putc v
    value = value[1].value
  end
end

Null = [].const
# Newline = [10.const,Null].const
# Space = [32.const,Null].const

def str_to_lazy_list(s,rhs=Null)
  to_lazy_list(s.chars.map(&:ord), rhs)
end

def to_lazy_list(l, rhs=Null, ind=0)
  ind >= l.size ? rhs.value : [l[ind].const, Promise.new{to_lazy_list(l, rhs, ind+1)}]
end

def to_eager_str(v)
  to_strict_list(v.const).map{|i|(i.to_i%256).chr}.join
end

def is_digit(i)
  i>=48 && i<58
end

def split_non_digits(s)
  return [] if s.empty
  v,found,s2=read_num(s)
  return [] if !found
  [v.const,Promise.new{split_non_digits(s2)}]
end

def read_num(s)
  multiplier=1
  until s.empty || is_digit(s.value[0].value)
    if s.value[0].value == ?-.ord
      multiplier *= -1
    else
      multiplier = 1
    end
    s = s.value[1]
  end
  v = 0
  found_int = false
  until s.empty || !is_digit(s.value[0].value)
    found_int = true
    v = v*10+s.value[0].value-48
    s = s.value[1]
  end

  [multiplier * v, found_int, s]
end

def lines(s)
  return [] if s.empty

  after = Promise.new{lines(s.value[1])}
  if s.value[0].value == 10
    [Null, after]
  else
    [Promise.new{
      after.empty ? [s.value[0], Null] : [s.value[0], after.value[0]]
     },
     Promise.new{
      after.empty ? [] : after.value[1].value
     }]
   end
end

def str(i)
  to_lazy_list(i.value.to_s.bytes.to_a)
end

def take(n, a)
  return [] if n < 1 || a.empty
  [a.value[0], Promise.new{ take(n-1, a.value[1]) }]
end

def drop(n, a)
  while n>=1 && !a.empty
    n-=1
    a=a.value[1]
  end
  a.value
end

def len(a)
  return 0 if a.empty?
  return 1+len(a[1].value)
end

def sort_by(a,b,rank)
  raise IogiiError.new "cannot sortby negative rank" if rank < 0 # todo ideally this done statically
  fromby(sort(toby(a,b),rank,true))
end

def toby(a,b)
  return Null if a.empty || b.empty
  Promise.new{ [[a.value[0], b.value[0]], toby(a.value[1], b.value[1])] }
end

def fromby(a)
  return [] if a == []
  [Promise.new{a[0][0].value}, Promise.new{fromby(a[1].value)}]
end

# It would be very interesting and useful to design a more lazy sorting algorithm
# so that you can select ith element in O(n) total time after sorting a list.
def sort(a,rank,by=false)
  return [] if a.empty
  return a.value if a.value[1].empty
  n=len(a.value)
  left=take(n/2, a).const
  right=drop(n/2, a).const
  merge(sort(left,rank,by),sort(right,rank,by),rank,by)
end

def merge(a,b,rank,by)
  return b if a==[]
  return a if b==[]
  if (by ? spaceship(a[0][1], b[0][1], rank) : spaceship(a[0], b[0],rank)) <= 0
    [a[0], Promise.new{merge(a[1].value,b,rank,by)}]
  else
    [b[0], Promise.new{merge(a,b[1].value,rank,by)}]
  end
end

def spaceship(a,b,rank)
  if rank>0
    until a.empty && b.empty
      return -1 if a.empty
      return 1 if b.empty
      s0 = spaceship(a.value[0],b.value[0],rank-1)
      return s0 if s0 != 0
      a = a.value[1]
      b = b.value[1]
    end
    return 0
  else
    a.value<=>b.value
  end
end

def inspectv(t,rank,value,rhs=Null)
  if t == CharType && rank == 1
    ['"'.ord.const, Promise.new{
      concat_map(value,Promise.new{str_to_lazy_list('"',rhs)}){|v,r,first|
       str_to_lazy_list(escape_str_char(v.value),r)
      }
    }]
  elsif rank > 0 #List
    ["[".ord.const, Promise.new{
      concat_map(value,Promise.new{str_to_lazy_list("]",rhs)}){|v,r,first|
        first ?
          inspectv(t,rank-1,v,r) :
          [','.ord.const,Promise.new{inspectv(t,rank-1,v,r)}]
      }
    }]
  elsif t == EmptyType
    value.value # take the value so it can throw
    raise
  elsif t == IntType
      str_to_lazy_list(value.value.to_s,rhs)
  elsif t == CharType
    str_to_lazy_list(inspect_char(value.value),rhs)
  else
    raise
  end
end

def filter(a,b,bt,rank)
  raise IogiiError.new "cannot filter by negative rank" if rank < 0 # todo ideally this done statically
  return [] if a.empty || b.empty
  if truthy(bt,rank,b.value[0].value)
   [a.value[0],Promise.new{ filter(a.value[1],b.value[1],bt,rank) }]
  else
    filter(a.value[1],b.value[1],bt,rank)
  end
end

def take_while(a,b,bt,rank)
  raise IogiiError.new "cannot take while negative rank" if rank < 0 # todo ideally this done statically
  return [] if a.empty || b.empty
  if truthy(bt,rank,b.value[0].value)
   [a.value[0],Promise.new{ take_while(a.value[1],b.value[1],bt,rank) }]
  else
    []
  end
end

def drop_until(a,b,bt,rank)
  raise IogiiError.new "cannot drop until negative rank" if rank < 0 # todo ideally this done statically
  until a.empty || b.empty
    if truthy(bt,rank,b.value[0].value)
      return a.value
    else
      a = a.value[1]
      b = b.value[1]
    end
  end
  []
end

def strip(a,at,rank)
  raise IogiiError.new "cannot strip negative rank" if rank < 0 # todo ideally this done statically
  lstrip = drop_until(a,a,at,rank)
  b = reverse(lstrip.const).const
  reverse(drop_until(b,b,at,rank).const)
end

# these aren't as lazy as possible, but it gets to use hashes
def tally(a,h=Hash.new(-1))
  return [] if a.empty
  [(h[to_strict_list(a.value[0])]+=1).const, Promise.new{tally(a.value[1], h)}]
end
def isuniq(a,h=Hash.new(-1))
  return [] if a.empty
  [(h[to_strict_list(a.value[0])]+=1) == 0 ? 1.const : 0.const, Promise.new{isuniq(a.value[1], h)}]
end

def set_diff(a,b)
  counts = Hash.new(0)
  to_strict_list(b).each{|be| counts[be] += 1 }
  diff_helper(a, counts)
end

def diff_helper(a, counts)
  while !a.empty && counts[ae=to_strict_list(a.value[0])] > 0
    a = a.value[1]
    counts[ae] -= 1
  end
  return [] if a.empty
  [a.value[0], Promise.new{ diff_helper(a.value[1], counts) }]
end

def range(a,b)
  return [] if a>=b
  [a.const, Promise.new{range(a+1,b)}]
end

def range_from(a)
  [a.const, Promise.new{range_from(a+1)}]
end

def reshape(a,b)
  return [] if a.empty
  [Promise.new{ take(b,a) },
   Promise.new{ reshape(Promise.new{ drop(b,a) }, b) }]
end

# truncate as soon as encounter empty list
def trunc(a)
  a.empty || a.value[0].empty ? [] : [a.value[0], Promise.new{trunc(a.value[1])}]
end

def map(a,&b)
  a.empty ? [] : [Promise.new{b[a.value[0]]}, Promise.new{map(a.value[1],&b)}]
end

def zero(rank,type=nil)
  return [] if rank > 0
  raise IogiiError.new("use of default value for empty type") if type == EmptyType
  type == CharType ? 32 : 0
end

def concat(a)
  concat_map(a,Null){|i,r,first|append(i,r)}
end

def tails(a)
  a.empty ? [Null, Null] : [a, Promise.new{ tails(a.value[1]) }]
end

def transpose(a)
  return [] if a.empty
  return transpose(a.value[1]) if a.value[0].empty
  broken = trunc(a.value[1]).const
  hds = Promise.new{ map(broken){|v|v.value[0].value} }
  tls = Promise.new{ map(broken){|v|v.value[1].value} }
  [Promise.new{[a.value[0].value[0],hds]},
   Promise.new{transpose [a.value[0].value[1],tls].const}]
end

def to_strict_list(a)
  sofar=[]
  loop {
    a=a.value
    return a if !(Array===a)
    return sofar if a==[]
    sofar<<to_strict_list(a[0])
    a=a[1]
  }
end

def to_base(a,b)
  reverse(to_base_h(a,b,a<=>0).const)
end

def to_base_h(a,b,sign)
  return [] if a==0
  raise IogiiError.new "base 0", nil if b==0
  digit = a%b*sign
  [digit.const, Promise.new{to_base_h((a-digit*sign)/b,b,sign)}]
end

def from_base(a,b)
  x=0
  until a.empty?
    x=x*b+a[0].value
    a=a[1].value
  end
  x
end


ReadStdin = Promise.new{ read_stdin }
def read_stdin
  c=STDIN.getc
  if c.nil?
    []
  else
    [c.ord.const, Promise.new{ read_stdin }]
  end
end
